import json
import uuid

from realnet.core.provider import OrgsProvider
from realnet.core.type import Org, Account, Authenticator, Client, Endpoint
from .models import AccountType, Org as OrgModel, Account as AccountModel, Authenticator as AuthenticatorModel, Client as ClientModel, Item as ItemModel, OrgRoleType, Type as TypeModel, Acl,  AclType, VisibilityType, session as db
from .utility import item_model_to_item, get_types_by_name, type_model_to_type, get_derived_types
from typing import cast
from urllib.parse import unquote
import uuid

from sqlalchemy.sql import func, or_
from sqlalchemy.dialects.postgresql import ARRAY

class SqlOrgsProvider(OrgsProvider):

    def get_orgs(self):
            return [Org(org.id, org.name) for org in db.query(OrgModel).all()]

    def get_org_by_id(self,id):
        org = db.query(OrgModel).filter(OrgModel.id == id).first()
        if org:
            return Org(org.id, org.name)
        return None

    def get_org_by_name(self,name):
        org = db.query(OrgModel).filter(OrgModel.name == name).first()
        if org:
            return Org(org.id, org.name)
        return None

    def get_account_by_id(self,id):
        account = db.query(AccountModel).get(id)
        if account:
            return Account(account.id, account.username, Org(account.org.id, account.org.name), account.org_role_type)
        else:
            return None

    def check_password(self, org_id, account_id, password):
        account = db.query(AccountModel).filter(AccountModel.org_id == org_id, AccountModel.username == account_id).first()
        if account:
            if password and account.check_password(password):
                return Account(account.id, account.username, Org(account.org.id, account.org.name), account.org_role_type)
        return None

    def get_org_authenticators(self, org_id):
        org = db.query(OrgModel).filter(OrgModel.name == org_id).first()
        if not org:
            # is there a public org?
            org = db.query(OrgModel).filter(OrgModel.public == True).first()
        if org:
            return [Authenticator(a.id, a.name, a.org_id, a.get_url()) for a in db.query(AuthenticatorModel).filter(AuthenticatorModel.org_id == org.id).all()]
        return []

    def get_org_clients(self, org_id):
        return [Client( client.id, 
                        client.name, 
                        Org(client.org.id, client.org.name), 
                        client.attributes) for client in db.query(ClientModel).filter(ClientModel.org_id == org_id).all()]

    def get_org_client(self, org_id, client_id):
        return db.query(ClientModel).filter(or_(ClientModel.client_id == client_id, ClientModel.name == client_id),ClientModel.org_id == org_id).first()

    def get_public_apps(self, org_id):
        tbn = get_types_by_name(org_id)
        type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name == 'App', TypeModel.org_id == org_id).all()]
        derived_type_ids = get_derived_types(org_id, type_ids)
        apps = db.query(ItemModel).filter(ItemModel.acls.any(Acl.type == AclType.public), ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))), ItemModel.org_id == org_id).all()
        return [item_model_to_item(org_id, app, tbn) for app in apps]

    def get_public_types(self, org_id):
        tbn = get_types_by_name(org_id)
        return tbn.values()

    def get_public_forms(self, org_id):
        tbn = get_types_by_name(org_id)
        type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name == 'Form', TypeModel.org_id == org_id).all()]
        derived_type_ids = get_derived_types(org_id, type_ids)
        forms = db.query(ItemModel).filter(ItemModel.acls.any(Acl.type == AclType.public), ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))), ItemModel.org_id == org_id).all()
        return [item_model_to_item(org_id, form, tbn) for form in forms]
    
    def get_public_items(self, org_id, data):
        op_or = False

        type_names = data.get('type_names')
        if type_names:
            data['type_names'] = type_names
        else:
            types = data.get('types')
            if types:
                data['type_names'] = types
                type_names = types
        

        tags = data.get('tags')
        if tags:
            data['tags'] = tags

        status = data.get('status')
        if status:
            data['status'] = status

        name = data.get('name')
        if name:
            data['name'] = name
        
        parent_id = data.get('parent_id')
        if parent_id:
            data['parent_id'] = parent_id
        else:
            if 'any_level' in data and str(data['any_level']).lower() == 'true': 
                pass
            else:
                data['parent_id'] = None

        location = data.get('location')
        if location:
            data['location'] = location

        valid_from = data.get('valid_from')
        if valid_from:
            data['valid_from'] = valid_from

        valid_to = data.get('valid_to')
        if valid_to:
            data['valid_to'] = valid_to

        status = data.get('status')
        if status:
            data['status'] = status

        # TODO below

        keys = data.get('keys')
        if keys:
            data['keys'] = keys

        values = data.get('values')
        if values:
            data['values'] = values

        visibility = data.get('visibility')
        if visibility:
            data['visibility'] = visibility

        op = data.get('op')

        if op:
            if op.lower() == 'or':
                op_or = True

        conditions = []

        if type_names:
            type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name.in_(type_names), TypeModel.org_id == self.org_id).all()]
            derived_type_ids = get_derived_types(self.org_id, type_ids)
            conditions.append(ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))))

        if tags:
            if isinstance(tags, list):
                conditions.append(ItemModel.tags.contains(cast(ARRAY(db.String), tags)))
            else:
                conditions.append(ItemModel.tags.contains([tags]))
        
        if name:
            conditions.append(ItemModel.name.ilike('{}%'.format(unquote(str(name)))))
        
        if parent_id:
            conditions.append(ItemModel.parent_id == parent_id)
        else:
            if 'any_level' in data and str(data['any_level']).lower() == 'true': 
                pass
            else:
                conditions.append(ItemModel.parent_id == None)

        if location:
            loc_data = json.loads(location)
            if isinstance(loc_data,str):
                    loc_data = json.loads(loc_data)
            if loc_data['type'] == 'Point':
                item_location = 'SRID=4326;POINT({0} {1})'.format(loc_data['coordinates'][0], loc_data['coordinates'][1])
                range = (0.00001) * float(500) #converting from meters to lng/lat scale
                conditions.append(func.ST_DWithin(ItemModel.location, item_location, range))
            else:
                item_location = 'SRID=4326;POLYGON(('
                for ii in loc_data['coordinates'][0]:
                    item_location = item_location + '{0} {1},'.format(ii[0], ii[1])
                item_location = item_location[0:-1] + '))'
                conditions.append(func.ST_Within(ItemModel.location, item_location))

        if valid_from:
            conditions.append(ItemModel.valid_from >= valid_from)

        if valid_to:
            conditions.append(ItemModel.valid_to <= valid_to)

        if status:
            conditions.append(ItemModel.status == status)

        
        if keys and values:
            if len(keys) > 1 and op_or:
                subconditions = []
                for kv in zip(keys, values):
                    subconditions.append(ItemModel.attributes.op('->>')(kv[0]) == kv[1])
                
                conditions.append(or_(*subconditions))
            else:
                for kv in zip(keys, values):
                    # conditions.append(ItemModel.attributes[kv[0]].astext == kv[1])
                    conditions.append(ItemModel.attributes.op('->>')(kv[0]) == kv[1])

        if visibility:
            conditions.append(ItemModel.visibility == VisibilityType[visibility])

        result_item_models = []

        conditions.append(ItemModel.org_id == org_id)
        conditions.append(ItemModel.acls.any(Acl.type == AclType.public))
        result_item_models = db.query(ItemModel).filter(*conditions).all()
        
        tbn = get_types_by_name(org_id)
        items = [item_model_to_item(org_id, i, tbn, False) for i in result_item_models]
        return items

    def get_public_orgs(self):
        return [Org(org.id, org.name) for org in db.query(OrgModel).filter(OrgModel.public == True).all()]

    def get_public_item(self, id):
        item = db.query(ItemModel).filter(ItemModel.acls.any(Acl.type == AclType.public), ItemModel.id == id).first()
        if item:
            tbn = get_types_by_name(item.org_id)
            return item_model_to_item(item.org_id, item, tbn)
        return None
    
    def get_org_login(self, org_id):
        tbn = get_types_by_name(org_id)
        type_ids = [ti.id for ti in db.query(TypeModel).filter(TypeModel.name == 'Resource', TypeModel.org_id == org_id).all()]
        derived_type_ids = get_derived_types(org_id, type_ids)
        login = db.query(ItemModel).filter(ItemModel.name == 'login', ItemModel.acls.any(Acl.type == AclType.public), ItemModel.type_id.in_(list(set(type_ids + derived_type_ids))), ItemModel.org_id == org_id).first()
        if login:
            return item_model_to_item(org_id, login, tbn)
        
        return None
        
        return [item_model_to_item(org_id, login, tbn) for login in logins][0] if logins else None

    def get_account_by_username(self, org_id, username):
        account = db.query(AccountModel).filter(AccountModel.org_id == org_id, AccountModel.username == username).first()
        if account:
            return Account(account.id, account.username, Org(account.org.id, account.org.name), account.org_role_type)
        return None

    def create_account(self, org_id, type, username, email, password, org_role_type):
        account = AccountModel(  id=str( uuid.uuid4()), 
                                    type=AccountType[type.lower()],
                                    username=username, 
                                    email=email, 
                                    org_id=org_id,
                                    org_role_type=OrgRoleType[org_role_type.lower()])
        account.set_password(password)
        db.add(account)
        db.commit()
        return account
    
    def delete_account(self, org_id, account_id):
        account = db.query(AccountModel).filter(AccountModel.org_id == org_id, AccountModel.id == account_id).first()
        if account:
            db.delete(account)
            db.commit()


from flask import render_template, redirect, current_app
from realnet.core.type import Instance, Item
from realnet.resource.items.items import Items

class Accounts(Items):
    
    def item_from_account(self, account, account_type):
        instance = Instance(account.id, account_type, account.name)
        return Item(account.org.id, account.org.id, instance, account.id, account.name, dict(), [])
    
    def item_from_role(self, role, role_type):
        instance = Instance(role.id, role_type, role.name)
        return Item(role.org.id, role.org.id, instance, role.id, role.name, dict(), role.apps)
    
    def item_from_account_role(self, account_role, account_role_type):
        instance = Instance(account_role.id, account_role_type, account_role.name)
        return Item(account_role.org.id, account_role.org.id, instance, account_role.id, account_role.name, dict(), [])
    
    def item_from_account_group(self, account_group, account_group_type):
        instance = Instance(account_group.id, account_group_type, account_group.name)
        return Item(account_group.org.id, account_group.org.id, instance, account_group.id, account_group.name, dict(), [])
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        tbn = {t.name:t for t in module.get_types()}
        if path:
            view = args.get('view', ['roles'])[0]
            if view == 'roles':
                roles = module.get_account_roles(path)
                return [self.item_from_account_role(r, tbn['AccountRole']) for r in roles]
            elif view == 'groups':
                return [self.item_from_account_group(r, tbn['AccountGroup']) for r in module.get_account_groups(path)]
        else:
            account = module.get_account()
            return [self.item_from_account(r, tbn['Account']) for r in module.get_org_accounts(account.org.id)]
    
    def get_item(self, module, endpoint, account, args, path):
        contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
        account = contextProvider.get_account_by_id(path)
        tbn = {t.name:t for t in module.get_types()}
        if account:
            return self.item_from_account(account, tbn['Account'])

        return None
        
        
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args:
                if 'role_id' in args:
                    module.add_account_role(args['parent_id'], args['role_id'])
                    return redirect('/accounts/{}'.format(args['parent_id']))
                elif 'group_id' in args:
                    module.add_account_group(args['parent_id'], args['group_id'])
                    return redirect('/accounts/{}?active_view=groups'.format(args['parent_id']))
            elif args.get('password') != None and args.get('password') == args.get('repeat_password'):
                contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
                contextProvider.create_account(   
                                        account.org.id,
                                        'person',
                                        args.get('username'),
                                        args.get('email'),
                                        args.get('password'),
                                        args.get('org_role', 'visitor'))
                return redirect('/accounts')
            
        return self.render_item(module, endpoint, args, path, content_type)

    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'id' in args:
                active_view = args.get('active_view', 'roles')
                if active_view == 'roles':
                    module.remove_account_role(args['parent_id'], args['id'])
                elif active_view == 'groups':
                    module.remove_account_group(args['parent_id'], args['id'])
                return redirect('/accounts/{}'.format(args['parent_id']))
            elif 'id' in args:
                contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
                contextProvider.delete_account(account.org.id, args['id'])
                return redirect('/accounts')

        return self.render_item(module, endpoint, args, path, content_type)



from flask import render_template, redirect
from realnet.resource.items.items import Items
from realnet.core.type import Instance, Item

class Roles(Items):
    
    def item_from_role(self, role, role_type):
        instance = Instance(role.id, role_type, role.name)
        return Item(role.org.id, role.org.id, instance, role.id, role.name, dict(), role.apps)
    
    def item_from_role_app(self, role_app, role_app_type):
        instance = Instance(role_app.id, role_app_type, role_app.name)
        return Item(role_app.org_id, role_app.org_id, instance, role_app.id, role_app.name, dict(), [])

    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        tbn = {t.name:t for t in module.get_types()}
        
        if path:
            role = module.get_role(path)
            if role:
                return [self.item_from_role_app(ra, tbn['RoleApp']) for ra in role.apps]

        
        return [self.item_from_role(r, tbn['Role']) for r in module.get_roles(module)]

    def get_item(self, module, endpoint, account, args, path):
        role = module.get_role(path)
        tbn = {t.name:t for t in module.get_types()}
        if role:
            return self.item_from_role(role, tbn['Role'])

        return None

    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'app_id' in args:
                app_id = args['app_id']
                role_id = args['parent_id']
                module.add_role_app(role_id, app_id)

                return redirect('/roles/{}'.format(role_id))
            else:
                # tbn = {t.name:t for t in module.get_types()}
                type = module.get_type_by_name(args.get('type', 'Role'))
                role_object = module.create_role(**args)
                for instance in type.instances:
                    if instance.type.name == 'RoleApp':
                        module.add_role_app(role_object.id, instance.name)
                        # role = self.item_from_role(role_object, type)
                        # self.create_child_items(module, type, role)
                        # break
                
                # role = self.item_from_role(role_object, type)
                # self.create_child_items(module, type, role)
                        
                return redirect('/roles')

            
        return self.render_item(module, endpoint, args, path, content_type)

    def put(self, module, endpoint, args, path=None, content_type='text/html'):
        params = dict()
        if 'name' in args:
            params['name'] = args['name']
        if 'attributes' in args:
            params['attributes'] = args['attributes']
        module.update_item(args['id'], **params)
        del args['id']
        del args['edit']
        del args['item_id']
        return self.render_item(module, endpoint, args, path, content_type)

    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'id' in args:
                module.remove_role_app(args['parent_id'], args['id'])
                return redirect('/roles/{}'.format(args['parent_id']))
            elif 'id' in args:
                module.delete_role(args['id'])
            elif path:
                parts = path.split('/')
                if len(parts) == 3 and parts[1] == 'apps':
                    module.remove_role_app(parts[0], parts[2])
                    return redirect('/roles/{}'.format(parts[0]))

        return self.render_item(module, endpoint, args, path, content_type)


import uuid

from flask import render_template, redirect, current_app
from realnet.resource.items.items import Items
from realnet.core.type import Instance, Item

class Login(Items):

    def item_from_auth(self, auth, auth_type):
        instance = Instance(auth.id, auth_type, auth.name, {"url":auth.url})
        return Item(auth.id, auth.org_id, instance, auth.id, auth.name, dict(), [])
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        tbn = {t.name:t for t in module.get_types()}
        account = module.get_account()
        contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
        return []
    
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        return self.render_item(module, endpoint, args, path, content_type)
    
    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()

        return self.render_item(module, endpoint, args, path, content_type)



from flask import render_template, redirect
from realnet.resource.items.items import Items

class Views(Items):
    
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        item_id = args.get('parent_id')
        if item_id:
            item = module.get_item(item_id);
            if item:
                if module.can_account_write_item(account, item):
                    view = dict()
                    if 'name' in args:
                        view['name'] = args['name']
                    if 'attributes' in args:
                        view['attributes'] = args.get('attributes',dict())
                    if 'type' in args:
                        view['type'] = args.get('type')
                    
                    views = item.attributes.get('views', []) + [view]
                    params = {'attributes': {'views': views} }
                    module.update_item(item_id, **params)
                    return redirect('/items/{}'.format(item_id))
                    # return self.render_item(module, args, item_id, content_type)
            
        return self.render_item(module, endpoint, args, path, content_type)

    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        parts = path.split('/')
        if len(parts) == 3 and parts[1] == 'apps':
            module.remove_role_app(parts[0], parts[2])
        elif len(parts) == 1:
            module.delete_role(path)
        if 'id' in args:
            del args['id']
        if 'delete' in args:
            del args['delete']
        if 'item_id' in args:
            del args['item_id']
        return self.render_item(module, endpoint, args, path, content_type)


from flask import render_template, redirect
from realnet.resource.items.items import Items
from realnet.core.type import Instance, Item

class Groups(Items):

    def item_from_group(self, group, group_type):
        instance = Instance(group.id, group_type, group.name)
        return Item(group.org.id, group.org.id, instance, group.id, group.name, dict(), [])
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        tbn = {t.name:t for t in module.get_types()}
        account = module.get_account()
        groups = module.get_org_groups(account.org.id)
        return [self.item_from_group(r, tbn['Group']) for r in groups]
    
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'account_id' in args:
                app_id = args['app_id']
                role_id = args['parent_id']
                module.add_role_app(role_id, app_id)

                return redirect('/groups/{}'.format(role_id))
            else:
                group = module.create_org_group(account.org.id, args.get('name'))
                return redirect('/groups')
            
        return self.render_item(module, endpoint, args, path, content_type)
    
    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'id' in args:
                pass
                # module.remove_role_app(args['parent_id'], args['id'])
                # return redirect('/roles/{}'.format(args['parent_id']))
            elif 'id' in args:
                module.remove_org_group(account.org.id, args['id'])
            elif 'name' in args:
                module.remove_org_group(account.org.id, args['name'])

        return self.render_item(module, endpoint, args, path, content_type)


import uuid
import base64
import json
from flask import render_template, jsonify, Response, redirect
from realnet.core.type import Resource, Item, Instance
from realnet.core.hierarchy import items_from_attributes

class Items(Resource):

    def render_items_json(self, module, endpoint, args, path):
        account = module.get_account()
        query = self.get_query(module, args, path)
        if path:
            item = self.get_item(module, endpoint, account, args, path)
            if item:
                return json.dumps(item.to_dict())
            else:
                return {}
        else:        
            items = self.get_items(module, endpoint, args, path, account, query, None)
            return json.dumps([item.to_dict() for item in items])

    def render_items_html(self, module, endpoint, args, path=None):
        return render_template(self.get_template(module, endpoint, args, path), 
                               **self.get_template_args(module, endpoint, args, path))

    def render_item(self, module, endpoint, args, path=None, content_type='text/html'):
        if path and path.endswith('/data'):
            parts = path.split('/')
            id = parts[:1][0]
            item = module.get_item(id, True)
            if not item:
                instance = module.get_instance_by_id(id)
                if not instance:
                    type = module.get_type_by_id(id)
                    item = type
                    output = module.get_data(id)
                else:
                    item = instance
                    output = module.get_data(id)
            else:
                output = module.get_data(id)

            if not output and item.instance:
                output = module.get_data(item.instance.id)
                if not output and item.instance.type:
                    output = module.get_data(item.instance.type.id)

            if output:
                return Response(
                            # base64.b64encode(output.bytes).decode('utf-8'),
                            output.bytes,
                            mimetype=output.mimetype,
                            headers={"Content-Disposition": "attachment;filename={}".format(item.name)})
                    
            return None
        
        if content_type == 'application/json':
            return self.render_items_json(module, endpoint, args, path)
        else:
            return self.render_items_html(module, endpoint, args, path)
        

    def get(self, module, endpoint, args, path=None, content_type='text/html'):
        return self.render_item(module, endpoint, args, path, content_type)

    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        if 'type' in args:
            if args['type'] == 'Attribute':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id'])
                    if item:
                        if module.can_account_write_item(account, item):
                            params = dict()
                            current_attrs = dict(item.attributes)
                            current_attrs[attrs['name']] = attrs['value']
                            params['attributes'] = current_attrs
                            module.update_item(attrs['parent_id'], **params)
                    if 'active_view' in attrs:
                        return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                    else:
                        return redirect('/items/{}'.format(attrs['parent_id']))
                    
            elif args['type'].endswith('View'):
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            view = dict()
                            if 'name' in args:
                                view['name'] = args['name']
                            if 'attributes' in args:
                                view['attributes'] = args.get('attributes',dict())
                            if 'type' in args:
                                view['type'] = args.get('type')
                            
                            views = item.attributes.get('views', []) + [view]
                            current_attrs = dict(item.attributes)
                            current_attrs['views'] = views
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
            elif args['type'] == 'FormItem':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            form = dict()
                            form['type'] = 'FormItem'
                            form['attributes'] = dict()
                            if 'name' in args:
                                form['name'] = args['name']
                            if 'path' in args:
                                form['attributes']['path'] = args['path']
                            if 'form' in args:
                                form['attributes']['form'] = args['form']
                            
                            forms = item.attributes.get('forms', []) + [form]
                            current_attrs = dict(item.attributes)
                            current_attrs['forms'] = forms
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
            elif args['type'] == 'MenuItem':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            form = dict()
                            form['type'] = 'MenuItem'
                            form['attributes'] = dict()
                            if 'name' in args:
                                form['name'] = args['name']
                            if 'path' in args:
                                form['attributes']['path'] = args['path']
                            if 'form' in args:
                                form['attributes']['form'] = args['form']
                            if 'icon' in args:
                                form['attributes']['icon'] = args['icon']
                            
                            forms = item.attributes.get('menu', []) + [form]
                            current_attrs = dict(item.attributes)
                            current_attrs['menu'] = forms
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
            elif args['type'] == 'Query':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            query = dict()
                            if 'name' in args:
                                query['name'] = args['name']
                            if 'children' in args and args['children'] != '':
                                query['children'] = 'true' if args['children'] == 'on' or args['children'] == '1' or args['children'] == 'true' else 'false'
                            if 'any_level' in args and args['any_level'] != '':
                                query['any_level'] = 'true' if args['any_level'] == 'on' or args['any_level'] == '1' or args['any_level'] == 'true' else 'false'  
                            if 'types' in args:
                                query['types'] = args['types']
                            if 'tags' in args:
                                query['tags'] = args['tags']
                            
                            current_attrs = dict(item.attributes)
                            current_attrs['query'] = query
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
            elif 'parent_id' in args:
                attrs = dict(args)
                attrs['parent_id'] = args['parent_id']
                account = module.get_account()
                parent_item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                if module.can_account_write_item(account, parent_item):
                    item = module.create_item(**attrs)
                    if 'active_view' in attrs:
                        return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                    else:
                        return redirect('/items/{}'.format(attrs['parent_id']))
        
        attrs = dict()
        for k,v in args.items():
            if k not in set(['name', 'base', 'type', 'parent_id']):
                if not 'attributes' in attrs:
                    attrs['attributes'] = dict()
                attrs['attributes'][k] = v
            else:
                attrs[k] = v

        if not 'attributes' in attrs:
            attrs['attributes'] = dict()
        
        # if not 'resource' in attrs['attributes'] or not attrs['attributes']['resource']:
        #    attrs['attributes']['resource'] = 'types'

        if 'type' in attrs:
            type = module.get_type_by_name(attrs['type'])
            if type and 'resource' in type.attributes:
                resource_name = type.attributes['resource']
                func = module.get_resource_method(module, endpoint, resource_name, 'post')
                if func and resource_name != 'items':
                    tbn = {t.name:t for t in module.get_types()}
                    external_resource = func.invoke(module, endpoint, args, path, 'application/json')
                    return self.item_from_json(module, external_resource.json, tbn)
            
        parent_item = None

        if 'parent_id' in args:
            account = module.get_account()
            parent_item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
            if not parent_item or not module.can_account_write_item(account, parent_item):
                return redirect('/items/{}'.format(attrs['parent_id']))
        
        item = module.create_item(**attrs)

        if content_type == 'application/json':
            return json.dumps(item.to_dict())
        else:
            return self.render_item(module, endpoint, args, path, content_type)

        

    def put(self, module, endpoint, args, path=None, content_type='text/html'):
        
        if 'type' in args:
            if args['type'] == 'Attribute':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id'])
                    if item:
                        if module.can_account_write_item(account, item):
                            params = dict()
                            current_attrs = dict(item.attributes)
                            current_attrs[attrs['name']] = attrs['value']
                            params['attributes'] = current_attrs
                            module.update_item(attrs['parent_id'], **params)
                    if 'active_view' in attrs:
                        return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                    else:
                        return redirect('/items/{}'.format(attrs['parent_id']))
                    
            elif args['type'].endswith('View'):
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            view = dict()
                            if 'name' in args:
                                view['name'] = args['name']
                            if 'attributes' in args:
                                view['attributes'] = args.get('attributes',dict())
                            if 'type' in args:
                                view['type'] = args.get('type')
                            
                            views = item.attributes.get('views', []) + [view]
                            current_attrs = dict(item.attributes)
                            current_attrs['views'] = views
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
            elif args['type'] == 'FormItem':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            form = dict()
                            form['type'] = 'FormItem'
                            form['attributes'] = dict()
                            if 'name' in args:
                                form['name'] = args['name']
                            if 'path' in args:
                                form['attributes']['path'] = args['path']
                            if 'form' in args:
                                form['attributes']['form'] = args['form']
                            
                            forms = item.attributes.get('forms', []) + [form]
                            current_attrs = dict(item.attributes)
                            current_attrs['forms'] = forms
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
            elif args['type'] == 'MenuItem':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            form = dict()
                            form['type'] = 'MenuItem'
                            form['attributes'] = dict()
                            if 'name' in args:
                                form['name'] = args['name']
                            if 'path' in args:
                                form['attributes']['path'] = args['path']
                            if 'form' in args:
                                form['attributes']['form'] = args['form']
                            if 'icon' in args:
                                form['attributes']['icon'] = args['icon']
                            
                            forms = item.attributes.get('menu', []) + [form]
                            current_attrs = dict(item.attributes)
                            current_attrs['menu'] = forms
                            params = {'attributes': current_attrs }
                            module.update_item(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/items/{}'.format(attrs['parent_id']))
        else:
            attrs = dict()
            for k,v in args.items():
                if k not in set(['name', 'base']):
                    if not 'attributes' in attrs:
                        attrs['attributes'] = dict()
                    attrs['attributes'][k] = v
                else:
                    attrs[k] = v

            if not 'attributes' in attrs:
                attrs['attributes'] = dict()
            
            if not 'resource' in attrs['attributes'] or not attrs['attributes']['resource']:
                attrs['attributes']['resource'] = 'types'

            item = module.create_item(**attrs)
            
        if content_type == 'application/json':
            return jsonify(item.to_dict())
        else:
            return self.render_item(module, endpoint, args, path, content_type)

    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        attrs = dict(args)
        id = None
        if path:
            id = path
        elif 'id' in attrs:
            id = attrs['id']
        elif 'parent_id' in attrs:
            id = attrs['parent_id']

        if id:
            account = module.get_account()
            item = self.get_item(module, endpoint, account, attrs, id);
            if item:
                if module.can_account_write_item(account, item):
                    if 'type' in args:
                        if args['type'].endswith('View'):
                            if 'name' in args:
                                name = args['name'].lower()
                                views = [] 
                                for view in item.attributes.get('views', []):
                                    if view['name'].lower() != name:
                                        views.append(view)
                                current_attrs = dict(item.attributes)
                                current_attrs['views'] = views
                                params = {'attributes': current_attrs }
                                module.update_item(attrs['id'], **params)
                                if 'active_view' in attrs:
                                    return redirect('/items/{}?view={}'.format(attrs['id'], attrs['active_view']))
                                else:
                                    return redirect('/items/{}'.format(attrs['id']))
                        if args['type'] == 'FormItem':
                            if 'name' in args:
                                name = args['name'].lower()
                                forms = [] 
                                for form in item.attributes.get('forms', []):
                                    if form['name'].lower() != name:
                                        forms.append(form)
                                current_attrs = dict(item.attributes)
                                current_attrs['forms'] = forms
                                params = {'attributes': current_attrs }
                                module.update_item(id, **params)
                                if 'active_view' in attrs:
                                    return redirect('/items/{}?view={}'.format(id, attrs['active_view']))
                                else:
                                    return redirect('/items/{}'.format(id))
                        elif args['type'] == 'MenuItem':
                            if 'name' in args:
                                name = args['name'].lower()
                                forms = [] 
                                for form in item.attributes.get('menu', []):
                                    if form['name'].lower() != name:
                                        forms.append(form)
                                current_attrs = dict(item.attributes)
                                current_attrs['menu'] = forms
                                params = {'attributes': current_attrs }
                                module.update_item(id, **params)
                                if 'active_view' in attrs:
                                    return redirect('/items/{}?view={}'.format(id, attrs['active_view']))
                                else:
                                    return redirect('/items/{}'.format(id))
                        elif args['type'] == 'Attribute':
                            if 'name' in args:
                                attrs = dict(item.attributes)
                                if args['name'] in attrs:
                                    del attrs[args['name']]
                                    params = dict()
                                    params['attributes'] = attrs
                                    module.update_item(id, **params)
                                    if 'active_view' in args:
                                        return redirect('/items/{}?view={}'.format(id, args['active_view']))
                                    else:
                                        return redirect('/items/{}'.format(id))
                        else:
                            if id:
                                if attrs['id'] == attrs['parent_id']:
                                    module.delete_type(id)
                                    return redirect('/items')
                                else:
                                    module.delete_item(id)
                            if 'parent_id' in attrs:
                                if 'active_view' in attrs:
                                    return redirect('/items/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                                else:
                                    return redirect('/items/{}'.format(attrs['parent_id']))
                                    
        # module.delete_type(args['id'])
        return redirect('/items/{}'.format(id))

    def message(self, module, endpoint, args, path=None, content_type='text/html'):
        pass

    def run(self, module, endpoint, args, path=None, content_type='text/html'):
        pass

    def get_data(self, module, endpoint, args, path=None, content_type='text/html'):
        pass

    def update_data(self, module, endpoint, args, path=None, content_type='text/html'):
        pass

    def delete_data(self, module, endpoint, args, path=None, content_type='text/html'):
        pass

    def get_items_from_attributes(self, module, types_by_name, account, attributes, key=None):
        return items_from_attributes(types_by_name, account, attributes, key)

    def item_from_json(self, module, item_json, tbn):
        account = module.get_account()
        type = tbn.get(item_json['type'], 'Item')
        instance = Instance(type.id, type, item_json['name'], dict())
        return Item(account.id, account.org.id, instance, item_json.get('id', str(uuid.uuid4())), item_json.get('name', None), item_json.get('attributes', None))

    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        if query:
            results = []
            external_types = []
            internal_types = []
            types = []

            if 'types' in query:
                types = query['types']


            tbn = {t.name: t for t in module.get_types()}
            if isinstance(types, str):
                type = tbn.get(types,None)
                if type:
                    if type.attributes and 'resource' in type.attributes:
                        external_types.append(type)
                    else:
                        internal_types.append(type)
            else:
                for typename in types:
                    type = tbn.get(typename, None)
                    if type:
                        if type.attributes and 'resource' in type.attributes and type.attributes['resource'] != 'items':
                            external_types.append(type)
                        else:
                            internal_types.append(type)
            if external_types:
                external_query = dict()
                for k,v in query.items():
                    external_query[k] = v

                if parent_item and 'children' in external_query and external_query['children'] == 'true':
                    external_query['parent_id'] = parent_item.id
                
                for external_type in external_types:
                    func = module.get_resource_method(module, endpoint, external_type.attributes['resource'], 'get')
                    if func:
                        external_resources = func.invoke(module, endpoint, external_query, None, 'application/json')
                        if external_resources:
                            data = json.loads(external_resources)
                            if isinstance(data, list):
                                external_items = [self.item_from_json(module, er, tbn) for er in data]
                            else:
                                external_items = [self.item_from_json(module, data, tbn)]
                            results = results + [er for er in external_items]
                
                if internal_types:
                    external_query['types'] = [t.name for t in internal_types]
                    if parent_item and 'children' in query and query['children'] == 'true':
                        external_query['parent_id'] = parent_item.id
                    results = results +  [i for i in module.find_items(external_query)]
            
            if internal_types:
                if parent_item and 'children' in query and query['children'] == 'true':
                    query['parent_id'] = parent_item.id
                results = results + [i for i in module.find_items(query)]
            
            
            return results

        return []

    def get_item(self, module, endpoint, account, args, path):
        item = module.get_item(path)
        if item and module.can_account_read_item(account, item):
            return item

        return None

    def get_query(self, module, args, path):
        query = dict()
        
        if 'types' in args:
            if isinstance(args['types'], str):
                query['types'] = [args['types']]
            else:
                query['types'] = args['types']

        if 'types[]' in args:
            if isinstance(args['types[]'], str):
                query['types'] = [args['types[]']]
            else:
                query['types'] = args['types[]']

        if 'parent_id' in args and args['parent_id']:
            if isinstance(args['parent_id'], str):
                query['parent_id'] = args['parent_id']
            else:
                query['parent_id'] = args['parent_id'][0]

        if 'any_level' in args:
            if isinstance(args['any_level'], str):
                query['any_level'] = args['any_level']
            else:
                query['any_level'] = args['any_level'][0]

        if 'children' in args:
            if isinstance(args['children'], str):
                query['children'] = args['children']
            else:
                query['children'] = args['children'][0]

        if 'my_items' in args:
            if isinstance(args['my_items'], str):
                query['children'] = args['my_items']
            else:
                query['children'] = args['my_items'][0]
        
        return query

    def get_types(self, module, account, query, parent_item=None):
        return None

    def get_template(self, module, endpoint, args, path):
        return "item.html"

    def get_controls(self, module, items, tbn):
        controls = []
        account = module.get_account()
        for item in items:
            controls = controls + [Item(account.id, account.org.id, Instance(str(uuid.uuid4()), tbn[item['type']], item['name']), str(uuid.uuid4()), item['name'], item['attributes'] if 'attributes' in item else dict())]

        return controls

    def create_child_items(self, module, type, item):
        account = module.get_account()
        for instance in type.instances:
            params = dict(**instance.to_dict())
            if 'id' in params:
                existing_instance = module.get_instance_by_id(params['id'])
                if not existing_instance:
                    created_instance = module.create_instance(**params)
                    created_item = module.create_item(type_id=type.id, 
                                                    instance_id=created_instance.id, 
                                                    org_id=account.org.id, 
                                                    parent_id=item.id, 
                                                    name=created_instance.name)
                else:
                    created_item = module.create_item(type_id=type.id, 
                                                    instance_id=existing_instance.id, 
                                                    org_id=account.org.id, 
                                                    parent_id=item.id, 
                                                    name=existing_instance.name)
                    
    def get_template_args(self, module, endpoint, args, path):
        account = module.get_account()
        org = module.get_org()
        apps = module.get_apps(module)
        forms = []
        
        active_app_name = endpoint.item.name
        active_view_name = args.get('view')
        
        if active_view_name and not isinstance(active_view_name, str):
            active_view_name = active_view_name[0]
        
        active_subview_name = args.get('subview')

        if active_subview_name and not isinstance(active_subview_name, str):
            active_subview_name = active_subview_name[0]

        app = next((a for a in apps if a.name.lower() == endpoint.item.name), None)

        if not app:
            app = next((a for a in apps), None)
        
        if apps:
            if active_app_name:
                app = next((a for a in apps if a.name.lower() == active_app_name), app)

        path_item = app
        target_item = app

        path_segments = []

        path_item_id = None
        target_item_id = None

        if path:
            path_segments = path.split('/')

        if path_segments:
            path_item_id = path_segments[0]
            path_item =  self.get_item(module, endpoint, account, args, path_item_id)
            target_item = path_item
        
        if 'item_id' in args:
            target_item_id = args['item_id']
            if isinstance(args['item_id'], str):
                target_item_id = args['item_id']
            else:
                target_item_id = args['item_id'][0]
                
            target_item = self.get_item(module, endpoint, account, args, target_item_id)

        if path_item:
            app = path_item


        views = app.attributes.get('views',[])

        items = []
        query = app.attributes.get('query')
        types = app.attributes.get('types',[])
        
        menu = app.attributes.get('menu',[])

        form = None
        
        active_view = None

        if views:
            active_subview = None
            if not active_view_name:
                active_view = next((v for v in views), None)
                if active_view:
                    active_view_name = active_view['name'].lower()
            else:
                active_view = next((v for v in views if v['name'].lower() == active_view_name), None)
            
            if active_view and 'attributes' in active_view:
                view_query = active_view['attributes'].get('query')
                if view_query:
                    query = view_query

                view_menu = active_view['attributes'].get('menu')
                if view_menu:
                    menu = view_menu

                view_types = active_view['attributes'].get('types')
                if view_types:
                    types = view_types

                subviews = [i for i in active_view.get('items',[])]
                if not active_subview_name:
                    active_subview = next((v for v in subviews), None)
                    if active_subview:
                        active_subview_name = active_subview.name
                else:
                    active_subview = next((v for v in subviews if v.name == active_subview_name), None)

            if active_subview:
                subview_query = active_subview.attributes.get('query')
                if subview_query:
                    query = subview_query

                subview_types = active_subview.attributes.get('types')
                if subview_types:
                    types = subview_types

        items = []

        tbn = {t.name:t for t in module.get_types()}

        if target_item:
            if target_item.instance and target_item.instance.type.is_derived_from('Form'):
                attrs = target_item.attributes
                if 'controls' in attrs:
                    items = self.get_controls(module, attrs['controls'], tbn)
            else:
                items = self.get_items(module, endpoint, args, path, account, query, path_item)

        root_path = '/' + endpoint.item.name

        if path_item_id:
            root_path = root_path + '/' + path_item_id

        if active_view_name:
            root_path = '?view={}&'.format(active_view_name)
        else:
            root_path = root_path + '?'

        typenames = []
        all_types = [t for t in tbn.values()]

        if types:
            type_filter = set(types)
            types_by_id = {t.id:t for t in tbn.values()}
            type_ids = set([t.id for t in tbn.values() if t.name in type_filter])
            derived_types = module.get_derived_types(type_ids)
            typenames = types + [types_by_id[t].name for t in derived_types if t in types_by_id]

        menu_forms = set()

        if menu:
            for mi in menu:
                menu_forms.add(mi['attributes']['form'])

        forms = [f for f in module.find_items({'types': ['Form'], 'any_level': 'true'})]

        for form in forms:
            if 'controls' in form.attributes:
                ctrls = self.get_controls(module, form.attributes['controls'], tbn)
                # for ctrl in ctrls:
                #    if ctrl.attributes and 'controls' in ctrl.attributes:
                #        sub_ctrls = self.get_controls(module, ctrl.attributes['controls'], tbn)
                #        ctrl.attributes['cotnrols'] = sub_ctrls
                form.items = ctrls
        
        all_forms = {form.name:form for form in forms}

        return {'app': app,
                'item': target_item,
                'apps': apps, 
                'org': org,
                'account': account,
                'items': items, 
                'views': views, 
                'active_view_name': active_view_name, 
                'active_app_name': active_app_name, 
                'active_subview_name': active_subview_name, 
                'types': types,
                'menu': menu,
                'forms': forms,
                'typenames': typenames,
                'root_path': root_path,
                'all_types': all_types,
                'all_forms': all_forms,
                'endpoint': endpoint.item.name,
                'view': active_view,
                'get_items_from_attributes': lambda attrs, key: self.get_items_from_attributes(module, tbn, account, attrs, key)}


import uuid

from realnet.provider.sql.models import Type as TypeModel, Instance as InstanceModel, Item as ItemModel, Acl as AclModel, session as db
from realnet.core.type import Type, Instance, Item
from realnet.core.acl import AclType, Acl

def traverse_instance(instances, instance, parent_type_name):
        for inst in instance.get('instances', []):
            instances.append({ "instance": inst, "parent_type_name": parent_type_name})
            traverse_instance(instances, inst, inst.get('type'))

def collect_inheritance_hierarchy(children, td, instances):
    name = td.get('name')
    derived_types = children.get(name)
    if derived_types:
        for derived_type in derived_types:
            for instance in td.get('instances', []):
                instances.append({ "instance": instance, "parent_type_name": derived_type['name']})
                # logging.getLogger().info('*** {} is derived from {} ***'.format(derived_type['name'], name))
            collect_inheritance_hierarchy(children, derived_type, instances)



def collect_type_dependencies(type_model):
    deps = set()
    if type_model.base:
        deps.add(type_model.base.name)
    for instance in type_model.instances:
        deps.add(instance.type.name)
    return deps

def collect_type_defs(type_models, type_defs):
    for td in type_models:
        name = td.name
        deps = collect_type_dependencies(td)
        if not name in type_defs: 
            type_defs[name] = {'type': td, 'deps': deps}
            # child_types = td.types
            # if child_types:
            #     collect_type_defs(child_types, type_defs)

def build_type(type, existing_types_by_name):
    base_name = None
    base_type = None

    if type.base:
        base_name = type.base.name
    
    if base_name:
        base_type = existing_types_by_name.get(base_name)
    
    instances = []
    type_instances = [i for i in type.instances]
    for instance in type_instances:
        attributes = instance.attributes
        instance_type = existing_types_by_name[instance.type.name]
        instances.append(Instance(instance.id, instance_type, instance.name, attributes))
        
    existing_types_by_name[type.name] = Type(type.id, type.name, base_type, type.attributes, instances)


def build_type_tree(queue, built, existing_types_by_name):
    to_build = []
    on_hold = []

    for element in queue:
        if element['deps'] <= built:
            to_build.append(element)
        else:
            on_hold.append(element)

    for element in to_build:
        build_type(element['type'], existing_types_by_name)
        built.add(element['type'].name)

    if on_hold:
        build_type_tree(on_hold, built, existing_types_by_name)

def get_types_by_name(org_id):
    type_defs = dict()
    type_models = db.query(TypeModel).filter(TypeModel.org_id == org_id).all()
    collect_type_defs(type_models, type_defs)
    existing_types_by_name = dict()

    build_type_tree(type_defs.values(), set(existing_types_by_name.keys()), existing_types_by_name)
    
    return existing_types_by_name

def get_types_by_name1(org_id):
    type_models = db.query(TypeModel).filter(TypeModel.org_id == org_id).all()
    instance_models = db.query(InstanceModel).filter(InstanceModel.org_id == org_id).all()
    type_models_by_name = {tm.name:tm for tm in type_models}
    type_names_without_instances = set([tm.name for tm in type_models])
    type_instances = dict()
    for instance_model in instance_models:
        if not instance_model.parent_type_id in type_instances:
            type_instances[instance_model.parent_type_id] = [instance_model]
        else:
            type_instances[instance_model.parent_type_id].append(instance_model)
        if instance_model.type.name in type_names_without_instances:
            type_names_without_instances.remove(instance_model.type.name)

    tbn = dict()
    
    # for tm in type_names_without_instances:
    #    type_model_to_type(org_id, type_models_by_name[tm], type_models_by_name, tbn, type_instances)

    for tm in type_models:
        type_model_to_type(org_id, tm, type_models_by_name, tbn, type_instances)

    return tbn

def type_model_to_type(org_id, type_model, type_models_by_name, types_by_name, type_instances):
    base = type_model.base
    attributes = dict(type_model.attributes or dict())
    if base:
        if base.name in types_by_name:
            base = types_by_name[base.name]
        else:
            base = type_model_to_type(org_id, base, type_models_by_name, types_by_name, type_instances)
        attributes = base.attributes | attributes
    
    instance_models = type_instances.get(type_model.id, [])
    instances = [instance_model_to_instance(im, types_by_name) for im in instance_models]
    result_type = Type(type_model.id, type_model.name, base, attributes, instances, type_model.module)
    types_by_name[type_model.name] = result_type
    return result_type

def fill_derived_types(base_types, type_ids, results):
    for type_id in type_ids:
        derived_types = base_types.get(type_id)
        if derived_types:
            for derived_type in derived_types:
                results.add(derived_type)
                fill_derived_types(base_types, [derived_type], results)

def get_derived_types(org_id, type_ids):
    type_models = db.query(TypeModel).filter(TypeModel.org_id == org_id).all()
    base_types = dict()
    
    for tm in type_models:
        if tm.base_id:
            existing_base_type = base_types.get(tm.base_id)
            if existing_base_type:
                existing_base_type.append(tm.id)
            else:
                base_types[tm.base_id] = [tm.id]

    results = set()
    fill_derived_types(base_types, type_ids, results)

    return list(results)

def instance_model_to_instance(instance_model, types_by_name):
    type = types_by_name.get(instance_model.type.name)
    if not type:
        return None
    return Instance(instance_model.id, type, instance_model.name, instance_model.attributes)

def acl_model_to_acl(acl_model):
    return Acl(acl_model.type, acl_model.permission, acl_model.target_id, acl_model.org_id, acl_model.owner_id)

def item_model_to_item(org_id, item_model, types_by_name, children=True):
    type = types_by_name.get(item_model.type.name)
    if children:
        child_items = db.query(ItemModel).filter(ItemModel.parent_id == item_model.id, InstanceModel.org_id == org_id).all()
        return Item(item_model.owner_id,
                    item_model.org_id,
                    Instance(item_model.instance_id, type, item_model.name),
                    item_model.id, 
                    item_model.name, 
                    item_model.attributes, 
                    [item_model_to_item(org_id, ci, types_by_name) for ci in child_items], 
                    [acl_model_to_acl(acl) for acl in item_model.acls])
    else:
        return Item(item_model.owner_id,
                    item_model.org_id,
                    Instance(item_model.instance_id, type, item_model.name),
                    item_model.id, 
                    item_model.name, 
                    item_model.attributes, 
                    [], 
                    [acl_model_to_acl(acl) for acl in item_model.acls])

def get_type_attributes(item_type):
    attributes = item_type.attributes
    if item_type.base_id:
        base_attributes = get_type_attributes(item_type.base)
        if base_attributes and attributes:
            attributes = base_attributes | attributes
        elif base_attributes:
            attributes = base_attributes
    return attributes

def get_type_instances(type):
    instances = []
    if type.instances:
        instances.extend(type.instances)
    if type.base:
        instances.extend(get_type_instances(type.base))
    return instances

def build_item( item_id,
                instance,
                attributes,
                item_data,
                owner_id,
                org_id,
                parent_item_id=None):

    location = item_data.get('item_location')
    valid_from = item_data.get('item_valid_from')
    valid_to = item_data.get('item_valid_to')
    status = item_data.get('item_status')
    tags = item_data.get('item_tags')
    linked_item_id = item_data.get('item_linked_item_id')

    if not location:
        item = ItemModel( id=item_id,
                    name=instance.name,
                    attributes=attributes,
                    valid_from=valid_from,
                    valid_to=valid_to,
                    status=status,
                    tags=tags,
                    owner_id=owner_id,
                    org_id=org_id,
                    instance_id=instance.id,
                    type_id=instance.type.id,
                    parent_id=parent_item_id,
                    linked_item_id=linked_item_id)
    else:
        item = ItemModel( id=item_id,
                    name=instance.name,
                    attributes=attributes,
                    valid_from=valid_from,
                    valid_to=valid_to,
                    status=status,
                    tags=tags,
                    owner_id=owner_id,
                    org_id=org_id,
                    location=location,
                    instance_id=instance.id,
                    type_id=instance.type.id,
                    parent_id=parent_item_id,
                    linked_item_id=linked_item_id)
    
    db.add(item)
    db.commit()

    create_public_acl = instance.public

    if item_data and 'item_is_public' in item_data:
        if str(item_data['item_is_public']).lower() == 'true':
            create_public_acl = True

    if create_public_acl:
        acl = AclModel(id=str(uuid.uuid4()),type='public', org_id=org_id, item_id=item_id)
        db.add(acl)
        db.commit()
    else:
        acl = AclModel(id=str(uuid.uuid4()),type='org', org_id=org_id, item_id=item_id, permission='r')
        db.add(acl)
        db.commit()
    
    # instances1 = instance.type.instances
    # instances2 = instance.type.base.instances
    instances = get_type_instances(instance.type)
    for child_instance in instances:
        attributes = get_type_attributes(child_instance.type)
        if attributes:
            if child_instance.attributes:
                attributes =  attributes | child_instance.attributes
        elif child_instance.attributes:
            attributes = child_instance.attributes
        
        child_item = build_item(  str(uuid.uuid4()),
                                  child_instance,
                                  attributes,
                                  {},
                                  owner_id,
                                  org_id,
                                  item.id)

    return item

def create_item_model(  db,
                        item_id,
                        item_instance_id,
                        item_type_name, 
                        item_name,
                        item_attributes,
                        item_location,
                        item_visibility,
                        item_tags,
                        item_is_public,
                        owner_id,
                        org_id,
                        parent_item_id=None,
                        item_valid_from=None,
                        item_valid_to=None,
                        item_status=None,
                        item_linked_item_id=None):

    item = None
    item_instance = db.query(InstanceModel).filter(InstanceModel.id == item_instance_id, InstanceModel.org_id == org_id).first()
    if not item_instance:
        item_type = db.query(TypeModel).filter(TypeModel.name == item_type_name, TypeModel.org_id == org_id).first()
    
        if not item_type:
            item_type = TypeModel(id=str(uuid.uuid4()),
                            name=item_type_name,
                            owner_id=owner_id,
                            org_id=org_id)
            db.add(item_type)
            db.commit()
        
        item_instance = InstanceModel(id=str(uuid.uuid4()),
                                      name=item_name,
                                      attributes=item_attributes,
                                      owner_id=owner_id,
                                      org_id=org_id,
                                      type_id=item_type.id)
        db.add(item_instance)
        db.commit()
    else:
        item_type = item_instance.type

    if item_type:
        # attributes = item_attributes | item_type.attributes
        attributes = get_type_attributes(item_type)
        
        if attributes:
            instance_attributes = item_instance.attributes
            if instance_attributes:
                attributes = attributes | instance_attributes
            if item_attributes:
                attributes =  attributes | item_attributes
        else:
            attributes = item_instance.attributes
            if attributes:
                if item_attributes:
                    attributes = attributes | item_attributes
            else:
                attributes = item_attributes 
        
        item_data = {"item_location": item_location, 
                     "item_visibility": item_visibility,
                     "item_tags": item_tags,
                     "item_is_public": item_is_public,
                     "item_valid_from": item_valid_from,
                     "item_valid_to": item_valid_to,
                     "item_status": item_status,
                     "item_linked_item_id": item_linked_item_id}
        
        item = build_item(item_id, item_instance, attributes, item_data, owner_id, org_id, parent_item_id)
    
    return item

def create_type_model(  db,
                        type_id,
                        type_name, 
                        type_icon,
                        type_attributes,
                        owner_id,
                        org_id,
                        type_module=None,
                        type_base_id=None):

    if type_base_id:
        type_base = db.query(TypeModel).filter(TypeModel.id == type_base_id, TypeModel.org_id == org_id).first()
        if not type_base:
            return None

    return TypeModel(id=type_id, name=type_name, icon=type_icon, attributes=type_attributes, module=type_module, base_id=type_base_id, owner_id=owner_id, org_id=org_id)


def create_instance_model(  instance_id,
                            instance_name,
                            instance_icon,
                            instance_attributes,
                            instance_public,
                            owner_id,
                            org_id,
                            instance_type_id,
                            instance_parent_type_id=None):
    return InstanceModel(id=instance_id,
                        name=instance_name,
                        icon=instance_icon,
                        attributes=instance_attributes,
                        public=instance_public,
                        owner_id=owner_id,
                        org_id=org_id,
                        type_id= instance_type_id,
                        parent_type_id=instance_parent_type_id)



from sqlalchemy import or_
from realnet.core.provider import RolesProvider
from realnet.core.type import Role, Org, Instance, Item
from .models import Role as RoleModel, Org as OrgModel, Item as ItemModel, RoleApp as RoleAppModel, Account as AccountModel, AccountRole as AccountRoleModel, session as db
from .utility import item_model_to_item, get_types_by_name
import uuid

class SqlRolesProvider(RolesProvider):

    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def item_from_role_app(self, role_app, link_type):
        instance = Instance(role_app.role_id, link_type, role_app.role.name)
        attributes = dict(role_app.app.attributes)
        attributes['resource'] = 'roles/{}/apps'.format(role_app.role_id)
        return Item(role_app.org_id, role_app.org_id, instance, role_app.app.id, role_app.app.name, attributes,[], [], role_app.app.id)

    def get_roles(self, module):
        tbn = get_types_by_name(self.org_id)
        return [Role(   role.id, 
                        role.name, 
                        Org(role.org.id, role.org.name), 
                        []) for role in db.query(RoleModel).filter(RoleModel.org_id == self.org_id).all()]

    def create_role(self, **kwargs):
        org = db.query(OrgModel).filter(OrgModel.id == self.org_id).first()
        args = dict(kwargs)
        role = RoleModel(id=str(uuid.uuid4()),name=args['name'], org_id=self.org_id)
        db.add(role)
        db.commit()
        return Role(role.id, role.name, org, [])
        

    def delete_role(self, id):
        role = db.query(RoleModel).filter(RoleModel.org_id == self.org_id, RoleModel.id == id).first()
        if role:
            db.delete(role)
            db.commit()

    def update_role(self, id, **kwargs):
        pass

    def get_role(self, id):
        tbn = get_types_by_name(self.org_id)
        role = db.query(RoleModel).filter(RoleModel.org_id == self.org_id, or_(RoleModel.id == id, RoleModel.name == id)).first()
        if role:
            role_apps = [role_app for role_app in role.apps]
            link_type = tbn['Link']
            items = []
            for role_app in role_apps:
                items.append(self.item_from_role_app(role_app, link_type))

            return Role(role.id, 
                        role.name, 
                        Org(role.org.id, role.org.name), 
                        items)
        return None

    def add_role_app(self, role_id, app_id):
        role = db.query(RoleModel).filter(RoleModel.org_id == self.org_id, RoleModel.id == role_id).first()
        app = db.query(ItemModel).filter(ItemModel.org_id == self.org_id, or_(ItemModel.id == app_id, ItemModel.name == app_id)).first()
        if role and app:
            role_app = db.query(RoleAppModel).filter(RoleAppModel.org_id == self.org_id, RoleAppModel.app_id == app.id, RoleAppModel.role_id == role_id).first()
            if not role_app:
                role_app = RoleAppModel(id=str(uuid.uuid4()),role_id=role_id, app_id=app.id, org_id=self.org_id)
                db.add(role_app)
                db.commit()
                
            return role_app.id

        return None

    def remove_role_app(self, role_id, app_id):
        role = db.query(RoleModel).filter(RoleModel.org_id == self.org_id, RoleModel.id == role_id).first()
        app = db.query(ItemModel).filter(ItemModel.org_id == self.org_id, or_(ItemModel.id == app_id, ItemModel.name == app_id)).first()
        if role and app:
            role_app = db.query(RoleAppModel).filter(RoleAppModel.org_id == self.org_id, RoleAppModel.app_id == app.id, RoleAppModel.role_id == role_id).first()
            if role_app:
                db.delete(role_app)
                db.commit()

    def add_account_role(self, account_id, role_id):
        role = db.query(RoleModel).filter(RoleModel.org_id == self.org_id, or_(RoleModel.id == role_id, RoleModel.name == role_id)).first()
        account = db.query(AccountModel).filter(AccountModel.org_id == self.org_id, AccountModel.id == account_id).first()
        if role and account:
            ar = AccountRoleModel(id=str(uuid.uuid4()), org_id=self.org_id, account_id=account.id, role_id=role_id)
            db.add(ar)
            db.commit()

    def remove_account_role(self, account_id, role_id):
        account = db.query(AccountModel).filter(AccountModel.org_id == self.org_id, AccountModel.id == account_id).first()
        account_role = db.query(AccountRoleModel).filter(AccountRoleModel.org_id == self.org_id, AccountRoleModel.account_id == account_id, AccountRoleModel.role_id == role_id).first()
        if account and account_role:
            db.delete(account_role)
            db.commit()

    def get_account_roles(self, account_id):
        account = db.query(AccountModel).filter(AccountModel.org_id == self.org_id, AccountModel.id == account_id).first()
        if account:
            return [Role(account_role.role.id, 
                        account_role.role.name, 
                        Org(account_role.role.org.id, account_role.role.org.name), 
                        account_role.role.apps) for account_role in account.roles]
        else:
            return []


from pygltflib import GLTF2, Scene, Mesh, Primitive, Node, Buffer, BufferView, Accessor, ELEMENT_ARRAY_BUFFER, ARRAY_BUFFER, SCALAR, UNSIGNED_SHORT, FLOAT, VEC3
import numpy as np
import trimesh
import trimesh.transformations as tf
import json
import base64
from PIL import Image, ImageDraw, ImageFont

def get_item_nodes(item, gltf):
    node = Node()

    mesh = get_item_mesh(item,gltf)
    size = len(gltf.meshes)
    gltf.meshes.append(mesh)
    node.mesh = size
    return node

def get_item_tmat(item):
    return [1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1]

def get_text_mesh(text):
    # Create a Pillow image object with the text "Hello, World!"
    font = ImageFont.truetype("arial.ttf", 36)
    size = font.getsize(text)
    image = Image.new("L", size, 255)
    draw = ImageDraw.Draw(image)
    draw.text((0, 0), text, font=font, fill=0)

    # Convert the image to a 2D mesh
    mesh = trimesh.creation.annulus(inner_radius=0, outer_radius=size[0]/2.0, segments=64, r_min=0, r_max=size[0]/2.0, height=10.0)
    for i, face in enumerate(mesh.faces):
        face_uv = mesh.vertices[face] / float(size[0])
        # mesh.visual.uv[i] = face_uv[:, :2]
        mesh.visual.face_colors[i] = [255, 255, 255, 255]  # white

    # Extrude the mesh to create a 3D mesh
    return mesh
    # Visualize the mesh
    extruded_mesh.show()

def import_json(tgltf, buffers, scene, gltf):
    nodes = []
    buffer_count = len(gltf.buffers)
    node_count = len(gltf.nodes)
    mesh_count = len(gltf.meshes)
    buffer_view_count = len(gltf.bufferViews)
    
    if tgltf:
        for bufferView in tgltf['bufferViews']:
            bv = BufferView()
            
            if 'buffer' in bufferView:
                bv.buffer = int(bufferView['buffer']) + buffer_count

            if 'byteOffset' in bufferView:
                bv.byteOffset = bufferView['byteOffset']

            if 'byteLength' in bufferView:
                bv.byteLength = bufferView['byteLength']

            gltf.bufferViews.append(bv)

        for asset in tgltf['asset']:
            pass

        for accessor in tgltf['accessors']:
            bv = Accessor()

            if 'bufferView' in accessor:
                bv.bufferView = int(accessor['bufferView']) + buffer_count

            if 'componentType' in accessor:
                bv.componentType = accessor['componentType']

            if 'count' in accessor:
                bv.count = accessor['count']

            if 'max' in accessor:
                bv.max = accessor['max']

            if 'min' in accessor:
                bv.min = accessor['min']

            if 'type' in accessor:
                bv.type = accessor['type']

            if 'byteOffset' in accessor:
                bv.byteOffset = accessor['byteOffset']

            if 'normalized' in accessor:
                bv.normalized = accessor['normalized']
            
            gltf.accessors.append(bv)

        for buffer in tgltf['buffers']:
            b = Buffer()

            if 'byteLength' in buffer:
                b.byteLength = int(buffer['byteLength'])

            if 'uri' in buffer:
                b.uri = 'data:application/octet-stream;base64,' + base64.b64encode(buffers[buffer['uri']]).decode()

            gltf.buffers.append(b)

        for mesh in tgltf['meshes']:
            m = Mesh()

            if 'name' in mesh:
                m.name = mesh['name']
            
            if 'primitives' in mesh:
                for primitive in mesh['primitives']:
                    p = Primitive()

                    if 'indices' in primitive:
                        p.indices = int(primitive['indices']) + buffer_view_count
                    
                    if 'mode' in primitive:
                        p.mode = primitive['mode']
                    
                    if 'attributes' in primitive:
                        for k,v in primitive['attributes'].items():
                            if k == 'POSITION':
                                p.attributes.POSITION = v + buffer_view_count
                            elif k == 'COLOR_0':
                                p.attributes.COLOR_0 = v + buffer_view_count
                            else:
                                pass
                    
                    m.primitives.append(p)

            gltf.meshes.append(m)

        for node in tgltf['nodes']:
            n = Node()

            if 'matrix' in node:
                n.matrix = node['matrix']
            else:
                bbox_size = [5, 5, 5]
                n.translation =  [d for d in np.random.uniform(low=-np.array(bbox_size)/2, high=np.array(bbox_size)/2)]

            if 'name' in node:
                n.name = node['name']

            if 'children' in node:
                n.children = [c + node_count for c in node['children']]

            if 'mesh' in node:
                n.mesh = int(node['mesh']) + mesh_count

            gltf.nodes.append(n)
        
        for scn in tgltf['scenes']:
            if 'nodes' in scn:
                for node in scn['nodes']:
                    nodes.append(int(node) + node_count)
    
    return nodes

def render_item(module, scene, item, gltf):
    nodes = []
    buffers = dict()
    tgltf = {}
    #1. Check if attributes have model if yes load from there if not
    if 'model' in item.attributes:
        tgltf = item.attributes['model']
    #2. Check if items have index.gltf file and if yes load from there if not
    else:
        files = [i for i in item.items if i.name == 'index.gltf']
        if files:
            # here we need to get the gltf data of the item
            tgltf = json.loads(module.get_data(files[0].id))
        else:
    #3. Render as sphere
            tscene = trimesh.Scene()
            geom = trimesh.creation.icosphere(radius=0.25)
            # geom = get_text_mesh(item.name)
            geom.visual.face_colors = np.random.uniform(
                0, 1, (len(geom.faces), 3))
            transform = tf.translation_matrix([0.1, -0.1, 0.1 / 2])
            tscene.add_geometry(geom, transform=transform)
            res = trimesh.exchange.gltf.export_gltf(tscene)

            for file, data in res.items():
                if file.endswith('.gltf'):
                    tgltf = json.loads(data)
                elif file.endswith('.bin'):
                    buffers[file] = data

    return import_json(tgltf, buffers, scene, gltf)




def build_gltf_from_items(module, items):
    gltf = GLTF2()
    scene = Scene()
    gltf.scenes.append(scene)
    for item in items:
        nodes = render_item(module, scene, item, gltf)
        for node in nodes:
            scene.nodes.append(node)

    gltf.save('triangle.gltf')
    return gltf